@@ -5,7 +5,7 @@
 
     Data structures for message catalogs.
 
-    :copyright: (c) 2013 by the Babel Team.
+    :copyright: (c) 2013-2019 by the Babel Team.
     :license: BSD, see LICENSE for more details.
 """
 
@@ -13,22 +13,23 @@ import re
 import time
 
 from cgi import parse_header
+from collections import OrderedDict
 from datetime import datetime, time as time_
 from difflib import get_close_matches
 from email import message_from_string
 from copy import copy
 
 from babel import __version__ as VERSION
-from babel.core import Locale
+from babel.core import Locale, UnknownLocaleError
 from babel.dates import format_datetime
 from babel.messages.plurals import get_plural
-from babel.util import odict, distinct, LOCALTZ, FixedOffsetTimezone
-from babel._compat import string_types, number_types, PY2, cmp
+from babel.util import distinct, LOCALTZ, FixedOffsetTimezone
+from babel._compat import string_types, number_types, PY2, cmp, text_type, force_text
 
 __all__ = ['Message', 'Catalog', 'TranslationError']
 
 
-PYTHON_FORMAT = re.compile(r'''(?x)
+PYTHON_FORMAT = re.compile(r'''
     \%
         (?:\(([\w]*)\))?
         (
@@ -37,7 +38,7 @@ PYTHON_FORMAT = re.compile(r'''(?x)
             [hlL]?
         )
         ([diouxXeEfFgGcrs%])
-''')
+''', re.VERBOSE)
 
 
 def _parse_datetime_header(value):
@@ -83,7 +84,7 @@ class Message(object):
                    pluralizable messages
         :param string: the translated message string, or a
                        ``(singular, plural)`` tuple for pluralizable messages
-        :param locations: a sequence of ``(filenname, lineno)`` tuples
+        :param locations: a sequence of ``(filename, lineno)`` tuples
         :param flags: a set or sequence of flags
         :param auto_comments: a sequence of automatic comments for the message
         :param user_comments: a sequence of user comments for the message
@@ -116,21 +117,13 @@ class Message(object):
         return '<%s %r (flags: %r)>' % (type(self).__name__, self.id,
                                         list(self.flags))
 
-    def __cmp__(self, obj):
+    def __cmp__(self, other):
         """Compare Messages, taking into account plural ids"""
-        def values_to_compare():
-            if isinstance(obj, Message):
-                plural = self.pluralizable
-                obj_plural = obj.pluralizable
-                if plural and obj_plural:
-                    return self.id[0], obj.id[0]
-                elif plural:
-                    return self.id[0], obj.id
-                elif obj_plural:
-                    return self.id, obj.id[0]
-            return self.id, obj.id
-        this, other = values_to_compare()
-        return cmp(this, other)
+        def values_to_compare(obj):
+            if isinstance(obj, Message) and obj.pluralizable:
+                return obj.id[0], obj.context or ''
+            return obj.id, obj.context or ''
+        return cmp(values_to_compare(self), values_to_compare(other))
 
     def __gt__(self, other):
         return self.__cmp__(other) > 0
@@ -275,11 +268,9 @@ class Catalog(object):
         :param fuzzy: the fuzzy bit on the catalog header
         """
         self.domain = domain
-        if locale:
-            locale = Locale.parse(locale)
         self.locale = locale
         self._header_comment = header_comment
-        self._messages = odict()
+        self._messages = OrderedDict()
 
         self.project = project or 'PROJECT'
         self.version = version or 'VERSION'
@@ -305,10 +296,40 @@ class Catalog(object):
         self.revision_date = revision_date
         self.fuzzy = fuzzy
 
-        self.obsolete = odict()  # Dictionary of obsolete messages
+        self.obsolete = OrderedDict()  # Dictionary of obsolete messages
         self._num_plurals = None
         self._plural_expr = None
 
+    def _set_locale(self, locale):
+        if locale is None:
+            self._locale_identifier = None
+            self._locale = None
+            return
+
+        if isinstance(locale, Locale):
+            self._locale_identifier = text_type(locale)
+            self._locale = locale
+            return
+
+        if isinstance(locale, string_types):
+            self._locale_identifier = text_type(locale)
+            try:
+                self._locale = Locale.parse(locale)
+            except UnknownLocaleError:
+                self._locale = None
+            return
+
+        raise TypeError('`locale` must be a Locale, a locale identifier string, or None; got %r' % locale)
+
+    def _get_locale(self):
+        return self._locale
+
+    def _get_locale_identifier(self):
+        return self._locale_identifier
+
+    locale = property(_get_locale, _set_locale)
+    locale_identifier = property(_get_locale_identifier)
+
     def _get_header_comment(self):
         comment = self._header_comment
         year = datetime.now(LOCALTZ).strftime('%Y')
@@ -318,9 +339,9 @@ class Catalog(object):
                          .replace('VERSION', self.version) \
                          .replace('YEAR', year) \
                          .replace('ORGANIZATION', self.copyright_holder)
-        if self.locale:
-            comment = comment.replace('Translations template', '%s translations'
-                                      % self.locale.english_name)
+        locale_name = (self.locale.english_name if self.locale else self.locale_identifier)
+        if locale_name:
+            comment = comment.replace('Translations template', '%s translations' % locale_name)
         return comment
 
     def _set_header_comment(self, string):
@@ -374,12 +395,12 @@ class Catalog(object):
         else:
             headers.append(('PO-Revision-Date', self.revision_date))
         headers.append(('Last-Translator', self.last_translator))
-        if self.locale is not None:
-            headers.append(('Language', str(self.locale)))
-        if (self.locale is not None) and ('LANGUAGE' in self.language_team):
+        if self.locale_identifier:
+            headers.append(('Language', str(self.locale_identifier)))
+        if self.locale_identifier and ('LANGUAGE' in self.language_team):
             headers.append(('Language-Team',
-                           self.language_team.replace('LANGUAGE',
-                                                      str(self.locale))))
+                            self.language_team.replace('LANGUAGE',
+                                                       str(self.locale_identifier))))
         else:
             headers.append(('Language-Team', self.language_team))
         if self.locale is not None:
@@ -393,7 +414,8 @@ class Catalog(object):
 
     def _set_mime_headers(self, headers):
         for name, value in headers:
-            name = name.lower()
+            name = force_text(name.lower(), encoding=self.charset)
+            value = force_text(value, encoding=self.charset)
             if name == 'project-id-version':
                 parts = value.split(' ')
                 self.project = u' '.join(parts[:-1])
@@ -402,6 +424,9 @@ class Catalog(object):
                 self.msgid_bugs_address = value
             elif name == 'last-translator':
                 self.last_translator = value
+            elif name == 'language':
+                value = value.replace('-', '_')
+                self._set_locale(value)
             elif name == 'language-team':
                 self.language_team = value
             elif name == 'content-type':
@@ -477,7 +502,7 @@ class Catalog(object):
         >>> Catalog(locale='en').num_plurals
         2
         >>> Catalog(locale='ga').num_plurals
-        3
+        5
 
         :type: `int`"""
         if self._num_plurals is None:
@@ -494,7 +519,9 @@ class Catalog(object):
         >>> Catalog(locale='en').plural_expr
         '(n != 1)'
         >>> Catalog(locale='ga').plural_expr
-        '(n==1 ? 0 : n==2 ? 1 : 2)'
+        '(n==1 ? 0 : n==2 ? 1 : n>=3 && n<=6 ? 2 : n>=7 && n<=10 ? 3 : 4)'
+        >>> Catalog(locale='ding').plural_expr  # unknown locale
+        '(n != 1)'
 
         :type: `string_types`"""
         if self._plural_expr is None:
@@ -536,7 +563,7 @@ class Catalog(object):
             buf.append('%s: %s' % (name, value))
         flags = set()
         if self.fuzzy:
-            flags |= set(['fuzzy'])
+            flags |= {'fuzzy'}
         yield Message(u'', '\n'.join(buf), flags=flags)
         for key in self._messages:
             yield self._messages[key]
@@ -625,7 +652,7 @@ class Catalog(object):
                    pluralizable messages
         :param string: the translated message string, or a
                        ``(singular, plural)`` tuple for pluralizable messages
-        :param locations: a sequence of ``(filenname, lineno)`` tuples
+        :param locations: a sequence of ``(filename, lineno)`` tuples
         :param flags: a set or sequence of flags
         :param auto_comments: a sequence of automatic comments
         :param user_comments: a sequence of user comments
@@ -673,7 +700,7 @@ class Catalog(object):
         if key in self._messages:
             del self._messages[key]
 
-    def update(self, template, no_fuzzy_matching=False):
+    def update(self, template, no_fuzzy_matching=False, update_header_comment=False, keep_user_comments=True):
         """Update the catalog based on the given template catalog.
 
         >>> from babel.messages import Catalog
@@ -728,7 +755,7 @@ class Catalog(object):
         """
         messages = self._messages
         remaining = messages.copy()
-        self._messages = odict()
+        self._messages = OrderedDict()
 
         # Prepare for fuzzy matching
         fuzzy_candidates = []
@@ -753,6 +780,10 @@ class Catalog(object):
             else:
                 oldmsg = remaining.pop(oldkey, None)
             message.string = oldmsg.string
+
+            if keep_user_comments:
+                message.user_comments = list(distinct(oldmsg.user_comments))
+
             if isinstance(message.id, (list, tuple)):
                 if not isinstance(message.string, (list, tuple)):
                     fuzzy = True
@@ -767,7 +798,7 @@ class Catalog(object):
                 message.string = message.string[0]
             message.flags |= oldmsg.flags
             if fuzzy:
-                message.flags |= set([u'fuzzy'])
+                message.flags |= {u'fuzzy'}
             self[message.id] = message
 
         for message in template:
@@ -779,7 +810,7 @@ class Catalog(object):
                     if no_fuzzy_matching is False:
                         # do some fuzzy matching with difflib
                         if isinstance(key, tuple):
-                            matchkey = key[0] # just the msgid, no context
+                            matchkey = key[0]  # just the msgid, no context
                         else:
                             matchkey = key
                         matches = get_close_matches(matchkey.lower().strip(),
@@ -798,9 +829,10 @@ class Catalog(object):
             if no_fuzzy_matching or msgid not in fuzzy_matches:
                 self.obsolete[msgid] = remaining[msgid]
 
-        # Allow the updated catalog's header to be rewritten based on the
-        # template's header
-        self.header_comment = template.header_comment
+        if update_header_comment:
+            # Allow the updated catalog's header to be rewritten based on the
+            # template's header
+            self.header_comment = template.header_comment
 
         # Make updated catalog's POT-Creation-Date equal to the template
         # used to update the catalog
